@@ -11,8 +11,13 @@
 #include <sys/sendfile.h>
 #include <sched.h>
 #include <sys/capability.h>
-#include <sys/mount.h>
+#include <sys/ioctl.h>
 #include <elf.h>
+#include <linux/fiemap.h>
+#include <linux/fs.h>
+
+#include "tty.h"
+#include "stats.h"
 
 #ifndef SEEK_DATA
 #define SEEK_DATA 3
@@ -29,6 +34,8 @@
  * and checked.
  */
 #define BUILD_ID_MAP_SIZE 1048576
+#define ST_UNIT		  512
+#define EXTENT_MAX_COUNT  512
 
 #include "cr_options.h"
 #include "imgset.h"
@@ -79,7 +86,7 @@ static LIST_HEAD(ghost_files);
 /*
  * When opening remaps we first create a link on the remap
  * target, then open one, then unlink. In case the remap
- * source has more than one instance, these tree steps
+ * source has more than one instance, these three steps
  * should be serialized with each other.
  */
 static mutex_t *remap_open_lock;
@@ -218,6 +225,92 @@ static int copy_file_to_chunks(int fd, struct cr_img *img, size_t file_size)
 	return 0;
 }
 
+static int skip_outstanding(struct fiemap_extent *fe, size_t file_size)
+{
+	/* Skip outstanding extent */
+	if (fe->fe_logical > file_size)
+		return 1;
+
+	/* Skip outstanding part of the extent */
+	if (fe->fe_logical + fe->fe_length > file_size)
+		fe->fe_length = file_size - fe->fe_logical;
+	return 0;
+}
+
+static int copy_file_to_chunks_fiemap(int fd, struct cr_img *img, size_t file_size)
+{
+	GhostChunkEntry ce = GHOST_CHUNK_ENTRY__INIT;
+	struct fiemap *fiemap_buf;
+	struct fiemap_extent *ext_buf;
+	int ext_buf_size, fie_buf_size;
+	off_t pos = 0;
+	unsigned int i;
+	int ret = 0;
+	int exit_code = 0;
+
+	ext_buf_size = EXTENT_MAX_COUNT * sizeof(struct fiemap_extent);
+	fie_buf_size = sizeof(struct fiemap) + ext_buf_size;
+
+	fiemap_buf = xzalloc(fie_buf_size);
+	if (!fiemap_buf) {
+		pr_perror("Out of memory when allocating fiemap");
+		return -1;
+	}
+
+	ext_buf = fiemap_buf->fm_extents;
+	fiemap_buf->fm_length = FIEMAP_MAX_OFFSET;
+	fiemap_buf->fm_flags |= FIEMAP_FLAG_SYNC;
+	fiemap_buf->fm_extent_count = EXTENT_MAX_COUNT;
+
+	do {
+		fiemap_buf->fm_start = pos;
+		memzero(ext_buf, ext_buf_size);
+		ret = ioctl(fd, FS_IOC_FIEMAP, fiemap_buf);
+		if (ret < 0) {
+			if (errno == EOPNOTSUPP) {
+				exit_code = -EOPNOTSUPP;
+			} else {
+				exit_code = -1;
+				pr_perror("fiemap ioctl() failed");
+			}
+			goto out;
+		} else if (fiemap_buf->fm_mapped_extents == 0) {
+			goto out;
+		}
+
+		for (i = 0; i < fiemap_buf->fm_mapped_extents; i++) {
+			if (skip_outstanding(&fiemap_buf->fm_extents[i], file_size))
+				continue;
+
+			ce.len = fiemap_buf->fm_extents[i].fe_length;
+			ce.off = fiemap_buf->fm_extents[i].fe_logical;
+
+			if (pb_write_one(img, &ce, PB_GHOST_CHUNK)) {
+				exit_code = -1;
+				goto out;
+			}
+
+			if (copy_chunk_from_file(fd, img_raw_fd(img), ce.off, ce.len)) {
+				exit_code = -1;
+				goto out;
+			}
+
+			if (fiemap_buf->fm_extents[i].fe_flags & FIEMAP_EXTENT_LAST) {
+				/* there are no extents left, break. */
+				goto out;
+			}
+		}
+
+		/* Record file's logical offset as pos */
+		pos = ce.len + ce.off;
+
+		/* Since there are still extents left, continue. */
+	} while (fiemap_buf->fm_mapped_extents == EXTENT_MAX_COUNT);
+out:
+	xfree(fiemap_buf);
+	return exit_code;
+}
+
 static int copy_chunk_to_file(int img, int fd, off_t off, size_t len)
 {
 	int ret;
@@ -314,46 +407,24 @@ static int mklnk_ghost(char *path, GhostFileEntry *gfe)
 static int ghost_apply_metadata(const char *path, GhostFileEntry *gfe)
 {
 	struct timeval tv[2];
-	int ret = -1;
 
-	if (S_ISLNK(gfe->mode)) {
-		if (lchown(path, gfe->uid, gfe->gid) < 0) {
-			pr_perror("Can't reset user/group on ghost %s", path);
-			goto err;
-		}
+	if (cr_fchpermat(AT_FDCWD, path, gfe->uid, gfe->gid, gfe->mode, AT_SYMLINK_NOFOLLOW) < 0)
+		return -1;
 
-		/*
-		 * We have no lchmod() function, and fchmod() will fail on
-		 * O_PATH | O_NOFOLLOW fd. Yes, we have fchmodat()
-		 * function and flag AT_SYMLINK_NOFOLLOW described in
-		 * man 2 fchmodat, but it is not currently implemented. %)
-		 */
-	} else {
-		if (chown(path, gfe->uid, gfe->gid) < 0) {
-			pr_perror("Can't reset user/group on ghost %s", path);
-			goto err;
-		}
+	if (!gfe->atim)
+		return 0;
 
-		if (chmod(path, gfe->mode)) {
-			pr_perror("Can't set perms %o on ghost %s", gfe->mode, path);
-			goto err;
-		}
+	tv[0].tv_sec = gfe->atim->tv_sec;
+	tv[0].tv_usec = gfe->atim->tv_usec;
+	tv[1].tv_sec = gfe->mtim->tv_sec;
+	tv[1].tv_usec = gfe->mtim->tv_usec;
+
+	if (lutimes(path, tv)) {
+		pr_perror("Can't set access and modification times on ghost %s", path);
+		return -1;
 	}
 
-	if (gfe->atim) {
-		tv[0].tv_sec = gfe->atim->tv_sec;
-		tv[0].tv_usec = gfe->atim->tv_usec;
-		tv[1].tv_sec = gfe->mtim->tv_sec;
-		tv[1].tv_usec = gfe->mtim->tv_usec;
-		if (lutimes(path, tv)) {
-			pr_perror("Can't set access and modification times on ghost %s", path);
-			goto err;
-		}
-	}
-
-	ret = 0;
-err:
-	return ret;
+	return 0;
 }
 
 static int create_ghost_dentry(char *path, GhostFileEntry *gfe, struct cr_img *img)
@@ -414,7 +485,7 @@ static int nomntns_create_ghost(struct ghost_file *gf, GhostFileEntry *gfe, stru
 	if (ghost_apply_metadata(path, gfe))
 		return -1;
 
-	strlcpy(gf->remap.rpath, path + 1, PATH_MAX);
+	__strlcpy(gf->remap.rpath, path + 1, PATH_MAX);
 	pr_debug("Remap rpath is %s\n", gf->remap.rpath);
 	return 0;
 }
@@ -545,7 +616,7 @@ static int open_remap_ghost(struct reg_file_info *rfi, RemapFilePathEntry *rpe)
 	gf->remap.rmnt_id = rfi->rfe->mnt_id;
 
 	if (S_ISDIR(gfe->mode))
-		strlcpy(gf->remap.rpath, rfi->path, PATH_MAX);
+		__strlcpy(gf->remap.rpath, rfi->path, PATH_MAX);
 	else
 		ghost_path(gf->remap.rpath, PATH_MAX, rfi, rpe);
 
@@ -910,10 +981,20 @@ static int dump_ghost_file(int _fd, u32 id, const struct stat *st, dev_t phys_de
 			goto err_out;
 		}
 
-		if (gfe.chunks)
-			ret = copy_file_to_chunks(fd, img, st->st_size);
-		else
+		if (gfe.chunks) {
+			if (opts.ghost_fiemap) {
+				ret = copy_file_to_chunks_fiemap(fd, img, st->st_size);
+				if (ret == -EOPNOTSUPP) {
+					pr_debug("file system don't support fiemap\n");
+					ret = copy_file_to_chunks(fd, img, st->st_size);
+				}
+			} else {
+				ret = copy_file_to_chunks(fd, img, st->st_size);
+			}
+		} else {
 			ret = copy_file(fd, img_raw_fd(img), st->st_size);
+		}
+
 		close(fd);
 		if (ret)
 			goto err_out;
@@ -946,8 +1027,8 @@ static int dump_ghost_remap(char *path, const struct stat *st, int lfd, u32 id,
 
 	pr_info("Dumping ghost file for fd %d id %#x\n", lfd, id);
 
-	if (st->st_size > opts.ghost_limit) {
-		pr_err("Can't dump ghost file %s of %" PRIu64 " size, increase limit\n", path, st->st_size);
+	if (st->st_blocks * ST_UNIT > opts.ghost_limit) {
+		pr_err("Can't dump ghost file %s of %" PRIu64 " size, increase limit\n", path, st->st_blocks * ST_UNIT);
 		return -1;
 	}
 
@@ -1035,7 +1116,6 @@ static int create_link_remap(char *path, int len, int lfd, u32 *idp, struct ns_i
 	RegFileEntry rfe = REG_FILE_ENTRY__INIT;
 	FownEntry fwn = FOWN_ENTRY__INIT;
 	int mntns_root;
-	int ret;
 	const struct stat *ost = &parms->stat;
 
 	if (!opts.link_remap_ok) {
@@ -1074,19 +1154,18 @@ static int create_link_remap(char *path, int len, int lfd, u32 *idp, struct ns_i
 
 	mntns_root = mntns_get_root_fd(nsid);
 
-again:
-	ret = linkat_hard(lfd, "", mntns_root, link_name, ost->st_uid, ost->st_gid, AT_EMPTY_PATH);
-	if (ret < 0 && errno == ENOENT) {
+	while (linkat_hard(lfd, "", mntns_root, link_name, ost->st_uid, ost->st_gid, AT_EMPTY_PATH) < 0) {
+		if (errno != ENOENT) {
+			pr_perror("Can't link remap to %s", path);
+			return -1;
+		}
+
 		/* Use grand parent, if parent directory does not exist. */
 		if (trim_last_parent(link_name) < 0) {
 			pr_err("trim failed: @%s@\n", link_name);
 			check_overlayfs_fallback(path, parms, fallback);
 			return -1;
 		}
-		goto again;
-	} else if (ret < 0) {
-		pr_perror("Can't link remap to %s", path);
-		return -1;
 	}
 
 	if (note_link_remap(link_name, nsid))
@@ -1571,22 +1650,10 @@ static int get_build_id_64(Elf64_Ehdr *file_header, unsigned char **build_id, co
  */
 static int get_build_id(const int fd, const struct stat *fd_status, unsigned char **build_id)
 {
-	char buf[SELFMAG + 1];
-	void *start_addr;
+	char *start_addr;
 	size_t mapped_size;
 	int ret = -1;
 
-	if (read(fd, buf, SELFMAG + 1) != SELFMAG + 1)
-		return -1;
-
-	/*
-	 * The first 4 bytes contain a magic number identifying the file as an
-	 * ELF file. They should contain the characters ‘\x7f’, ‘E’, ‘L’, and
-	 * ‘F’, respectively. These characters are together defined as ELFMAG.
-	 */
-	if (strncmp(buf, ELFMAG, SELFMAG))
-		return -1;
-
 	/*
 	 * If the build-id exists, then it will most likely be present in the
 	 * beginning of the file. Therefore at most only the first 1 MB of the
@@ -1594,16 +1661,25 @@ static int get_build_id(const int fd, const struct stat *fd_status, unsigned cha
 	 */
 	mapped_size = min_t(size_t, fd_status->st_size, BUILD_ID_MAP_SIZE);
 	start_addr = mmap(0, mapped_size, PROT_READ, MAP_PRIVATE | MAP_FILE, fd, 0);
-	if (start_addr == MAP_FAILED) {
+	if ((void*)start_addr == MAP_FAILED) {
 		pr_warn("Couldn't mmap file with fd %d\n", fd);
 		return -1;
 	}
 
-	if (buf[EI_CLASS] == ELFCLASS32)
-		ret = get_build_id_32(start_addr, build_id, fd, mapped_size);
-	if (buf[EI_CLASS] == ELFCLASS64)
-		ret = get_build_id_64(start_addr, build_id, fd, mapped_size);
+	/*
+	 * The first 4 bytes contain a magic number identifying the file as an
+	 * ELF file. They should contain the characters ‘\x7f’, ‘E’, ‘L’, and
+	 * ‘F’, respectively. These characters are together defined as ELFMAG.
+	 */
+	if (memcmp(start_addr, ELFMAG, SELFMAG))
+		goto out;
 
+	if (start_addr[EI_CLASS] == ELFCLASS32)
+		ret = get_build_id_32((Elf32_Ehdr *)start_addr, build_id, fd, mapped_size);
+	if (start_addr[EI_CLASS] == ELFCLASS64)
+		ret = get_build_id_64((Elf64_Ehdr *)start_addr, build_id, fd, mapped_size);
+
+out:
 	munmap(start_addr, mapped_size);
 	return ret;
 }
@@ -1688,6 +1764,7 @@ int dump_one_reg_file(int lfd, u32 id, const struct fd_parms *p)
 	int ret;
 	FileEntry fe = FILE_ENTRY__INIT;
 	RegFileEntry rfe = REG_FILE_ENTRY__INIT;
+	bool skip_for_shell_job = false;
 
 	if (!p->link) {
 		if (fill_fdlink(lfd, p, &_link))
@@ -1715,7 +1792,17 @@ int dump_one_reg_file(int lfd, u32 id, const struct fd_parms *p)
 	}
 
 	if (mi && mnt_is_overmounted(mi)) {
-		pr_err("Open files on overmounted mounts are not supported yet\n");
+		if (opts.shell_job && is_tty(p->stat.st_rdev, p->stat.st_dev)) {
+			skip_for_shell_job = true;
+		} else {
+			pr_err("Can't lookup mount=%d for fd=%d path=%s\n", p->mnt_id, p->fd, link->name + 1);
+			return -1;
+		}
+	}
+
+	if (!skip_for_shell_job && mnt_is_overmounted(mi)) {
+		pr_err("Open files on overmounted mounts are not supported yet; mount=%d fd=%d path=%s\n",
+		       p->mnt_id, p->fd, link->name + 1);
 		return -1;
 	}
 
@@ -1734,7 +1821,7 @@ int dump_one_reg_file(int lfd, u32 id, const struct fd_parms *p)
 		return -1;
 	}
 
-	if (mi && check_path_remap(link, p, lfd, id, mi->nsid))
+	if (mi && !skip_for_shell_job && check_path_remap(link, p, lfd, id, mi->nsid))
 		return -1;
 	rfe.name = &link->name[1];
 ext:
@@ -2202,9 +2289,21 @@ ext:
 		if (!validate_file(tmp, &st, rfi))
 			goto err;
 
-		if (rfi->rfe->has_mode && (st.st_mode != rfi->rfe->mode)) {
-			pr_err("File %s has bad mode 0%o (expect 0%o)\n", rfi->path, (int)st.st_mode, rfi->rfe->mode);
-			goto err;
+		if (rfi->rfe->has_mode) {
+			mode_t curr_mode = st.st_mode;
+			mode_t saved_mode = rfi->rfe->mode;
+
+			if (opts.skip_file_rwx_check) {
+				curr_mode &= ~(S_IRWXU | S_IRWXG | S_IRWXO);
+				saved_mode &= ~(S_IRWXU | S_IRWXG | S_IRWXO);
+			}
+
+			if (curr_mode != saved_mode) {
+				pr_err("File %s has bad mode 0%o (expect 0%o)\n"
+				       "File r/w/x checks can be skipped with the --skip-file-rwx-check option\n",
+				       rfi->path, (int)curr_mode, saved_mode);
+				goto err;
+			}
 		}
 
 		/*
@@ -2412,9 +2511,10 @@ static int open_filemap(int pid, struct vma_area *vma)
 			 * using dup because dup returns a reference to the same struct file inside kernel, but we
 			 * cannot open a new FD.
 			 */
-			ret = dup(plugin_fd);
+			ret = plugin_fd;
 		} else if (vma->e->status & VMA_AREA_MEMFD) {
-			ret = memfd_open(vma->vmfd, &flags);
+			if (!inherited_fd(vma->vmfd, &ret))
+				ret = memfd_open(vma->vmfd, &flags, true);
 		} else {
 			ret = open_path(vma->vmfd, do_open_reg_noseek_flags, &flags);
 		}
