@@ -22,6 +22,7 @@
 #include "rst-malloc.h"
 #include "log.h"
 #include "common/list.h"
+#include "util.h"
 #include "util-pie.h"
 #include "proc_parse.h"
 #include "file-ids.h"
@@ -402,8 +403,7 @@ static int tty_verify_active_pairs(void)
 {
 	unsigned long i, unpaired_slaves = 0;
 
-	for_each_bit(i, tty_active_pairs)
-	{
+	for_each_bit(i, tty_active_pairs) {
 		if ((i % 2) == 0) {
 			if (test_bit(i + 1, tty_active_pairs)) {
 				i++;
@@ -883,7 +883,7 @@ static int restore_tty_params(int fd, struct tty_info *info)
 	}
 
 	if (info->tie->has_uid && info->tie->has_gid) {
-		if (fchown(fd, info->tie->uid, info->tie->gid)) {
+		if (cr_fchown(fd, info->tie->uid, info->tie->gid)) {
 			pr_perror("Can't setup uid %d gid %d on %#x", (int)info->tie->uid, (int)info->tie->gid,
 				  info->tfe->id);
 			return -1;
